/*-*- mode:c;indent-tabs-mode:nil;c-basic-offset:2;tab-width:8;coding:utf-8 -*-│
│vi: set net ft=c ts=2 sts=2 sw=2 fenc=utf-8                                :vi│
╞══════════════════════════════════════════════════════════════════════════════╡
│ Copyright 2020 Justine Alexandra Roberts Tunney                              │
│                                                                              │
│ Permission to use, copy, modify, and/or distribute this software for         │
│ any purpose with or without fee is hereby granted, provided that the         │
│ above copyright notice and this permission notice appear in all copies.      │
│                                                                              │
│ THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL                │
│ WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED                │
│ WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE             │
│ AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL         │
│ DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR        │
│ PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER               │
│ TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR             │
│ PERFORMANCE OF THIS SOFTWARE.                                                │
╚─────────────────────────────────────────────────────────────────────────────*/
#include "libc/assert.h"
#include "libc/calls/struct/iovec.h"
#include "libc/runtime/zipos.internal.h"
#include "libc/sysv/consts/s.h"
#include "libc/sysv/errfuns.h"
#include "libc/zip.internal.h"

static size_t GetIovSize(const struct iovec *iov, size_t iovlen) {
  size_t i, r;
  for (r = i = 0; i < iovlen; ++i) r += iov[i].iov_len;
  return r;
}

static ssize_t __zipos_read_impl(struct ZiposHandle *h, const struct iovec *iov,
                                 size_t iovlen, ssize_t opt_offset) {
  int i;
  int64_t b, x, y;
  if (h->cfile == ZIPOS_SYNTHETIC_DIRECTORY ||
      S_ISDIR(GetZipCfileMode(h->zipos->map + h->cfile))) {
    return eisdir();
  }
  if (opt_offset == -1) {
    x = y = h->pos;
  } else {
    x = y = opt_offset;
  }
  for (i = 0; i < iovlen && y < h->size; ++i, y += b) {
    b = MIN(iov[i].iov_len, h->size - y);
    if (b) memcpy(iov[i].iov_base, h->mem + y, b);
  }
  if (opt_offset == -1) {
    h->pos = y;
  }
  return y - x;
}

/**
 * Reads data from zip store object.
 *
 * @return [1..size] bytes on success, 0 on EOF, or -1 w/ errno; with
 *     exception of size==0, in which case return zero means no error
 * @asyncsignalsafe
 */
ssize_t __zipos_read(struct ZiposHandle *h, const struct iovec *iov,
                     size_t iovlen, ssize_t opt_offset) {
  ssize_t rc;
  unassert(opt_offset >= 0 || opt_offset == -1);
  pthread_mutex_lock(&h->lock);
  rc = __zipos_read_impl(h, iov, iovlen, opt_offset);
  pthread_mutex_unlock(&h->lock);
  return rc;
}
